@@ -1,31 +1,54 @@
 from __future__ import annotations
 
-import glob
+import json
 import os
 import platform
 import shutil
 import subprocess
 import sys
 import sysconfig
-from distutils import log
-from distutils.errors import (
+import logging
+from setuptools.errors import (
     CompileError,
-    DistutilsExecError,
-    DistutilsFileError,
-    DistutilsPlatformError,
+    ExecError,
+    FileError,
+    PlatformError,
 )
-from distutils.sysconfig import get_config_var
-from typing import Dict, List, NamedTuple, Optional, Set, Tuple, cast
+from sysconfig import get_config_var
+from pathlib import Path
+from typing import Dict, List, Literal, NamedTuple, Optional, Set, Tuple, cast
 
-from setuptools.command.build import build as CommandBuild  # type: ignore[import]
+from setuptools import Distribution
+from setuptools.command.build import build as CommandBuild
 from setuptools.command.build_ext import build_ext as CommandBuildExt
 from setuptools.command.build_ext import get_abi3_suffix
-from typing_extensions import Literal
+from setuptools.command.install_scripts import install_scripts as CommandInstallScripts
 
 from ._utils import format_called_process_error
 from .command import RustCommand
 from .extension import Binding, RustBin, RustExtension, Strip
-from .rustc_info import get_rust_host, get_rust_target_list, get_rustc_cfgs
+from .rustc_info import (
+    get_rust_host,
+    get_rust_version,
+    get_rustc_cfgs,
+)
+
+logger = logging.getLogger(__name__)
+
+
+try:
+    from wheel.bdist_wheel import bdist_wheel as CommandBdistWheel
+except ImportError:  # wheel installation might be deferred in PEP 517
+    from setuptools import Command as CommandBdistWheel
+
+
+def _check_cargo_supports_crate_type_option() -> bool:
+    version = get_rust_version()
+
+    if version is None:
+        return False
+
+    return version.major > 1 or (version.major == 1 and version.minor >= 64)  # type: ignore
 
 
 class build_rust(RustCommand):
@@ -62,6 +85,7 @@ class build_rust(RustCommand):
         self.qbuild = None
         self.build_temp = None
         self.plat_name = None
+        self.build_number = None
         self.target = os.getenv("CARGO_BUILD_TARGET")
         self.cargo = os.getenv("CARGO", "cargo")
 
@@ -69,7 +93,7 @@ class build_rust(RustCommand):
         super().finalize_options()
 
         if self.plat_name is None:
-            self.plat_name = cast(  # type: ignore[no-any-unimported]
+            self.plat_name = cast(
                 CommandBuild, self.get_finalized_command("build")
             ).plat_name
             assert isinstance(self.plat_name, str)
@@ -82,6 +106,9 @@ class build_rust(RustCommand):
             ("inplace", "inplace"),
         )
 
+        if self.build_number is not None and not self.build_number[:1].isdigit():
+            raise ValueError("Build tag (build-number) must start with a digit.")
+
     def run_for_extension(self, ext: RustExtension) -> None:
         assert self.plat_name is not None
 
@@ -112,37 +139,26 @@ class build_rust(RustCommand):
     def build_extension(
         self, ext: RustExtension, forced_target_triple: Optional[str] = None
     ) -> List["_BuiltModule"]:
-
-        target_info = self._detect_rust_target(forced_target_triple)
-        if target_info is not None:
-            target_triple = target_info.triple
-            cross_lib = target_info.cross_lib
-            linker = target_info.linker
-            # We're ignoring target_info.linker_args for now because we're not
-            # sure if they will always do the right thing. Might help with some
-            # of the OS-specific logic if it does.
-
-        else:
-            target_triple = None
-            cross_lib = None
-            linker = None
-
+        target_triple = self._detect_rust_target(forced_target_triple)
         rustc_cfgs = get_rustc_cfgs(target_triple)
 
-        env = _prepare_build_environment(cross_lib)
+        env = _prepare_build_environment()
 
         if not os.path.exists(ext.path):
-            raise DistutilsFileError(
-                f"can't find Rust extension project file: {ext.path}"
+            raise FileError(
+                f"can't find manifest for Rust extension `{ext.name}` at path `{ext.path}`"
             )
 
         quiet = self.qbuild or ext.quiet
         debug = self._is_debug_build(ext)
+        use_cargo_crate_type = _check_cargo_supports_crate_type_option()
 
-        # Find where to put the temporary build files created by `cargo`
-        target_dir = _base_cargo_target_dir(ext, quiet=quiet)
-        if target_triple is not None:
-            target_dir = os.path.join(target_dir, target_triple)
+        package_id = ext.metadata(quiet=quiet)["resolve"]["root"]
+        if package_id is None:
+            raise FileError(
+                f"manifest for Rust extention `{ext.name}` at path `{ext.path}` is a virtual manifest (a workspace root without a package).\n\n"
+                "If you intended to build for a workspace member, set `path` for the extension to the member's Cargo.toml file."
+            )
 
         cargo_args = self._cargo_args(
             ext=ext, target_triple=target_triple, release=not debug, quiet=quiet
@@ -150,21 +166,32 @@ class build_rust(RustCommand):
 
         rustflags = []
 
-        if linker is not None:
-            rustflags.extend(["-C", "linker=" + linker])
-
         if ext._uses_exec_binding():
-            command = [self.cargo, "build", "--manifest-path", ext.path, *cargo_args]
-
-        else:
-            rustc_args = [
-                "--crate-type",
-                "cdylib",
-                *ext.rustc_flags,
+            command = [
+                self.cargo,
+                "build",
+                "--manifest-path",
+                ext.path,
+                "--message-format=json-render-diagnostics",
+                *cargo_args,
             ]
 
+        else:
+            # If toolchain >= 1.64.0, use '--crate-type' option of cargo.
+            # See https://github.com/PyO3/setuptools-rust/issues/320
+            if use_cargo_crate_type:
+                rustc_args = [
+                    *ext.rustc_flags,
+                ]
+            else:
+                rustc_args = [
+                    "--crate-type",
+                    "cdylib",
+                    *ext.rustc_flags,
+                ]
+
             # OSX requires special linker arguments
-            if sys.platform == "darwin":
+            if rustc_cfgs.get("target_os") == "macos":
                 ext_basename = os.path.basename(self.get_dylib_ext_path(ext, ext.name))
                 rustc_args.extend(
                     [
@@ -180,10 +207,20 @@ class build_rust(RustCommand):
                 # the cdylib, see https://github.com/rust-lang/cargo/issues/10143
                 rustflags.append("-Ctarget-feature=-crt-static")
 
+            elif (rustc_cfgs.get("target_arch"), rustc_cfgs.get("target_os")) == (
+                "wasm32",
+                "emscripten",
+            ):
+                rustc_args.extend(["-C", "link-args=-sSIDE_MODULE=2 -sWASM_BIGINT"])
+
+            if use_cargo_crate_type and "--crate-type" not in cargo_args:
+                cargo_args.extend(["--crate-type", "cdylib"])
+
             command = [
                 self.cargo,
                 "rustc",
                 "--lib",
+                "--message-format=json-render-diagnostics",
                 "--manifest-path",
                 ext.path,
                 *cargo_args,
@@ -207,78 +244,84 @@ class build_rust(RustCommand):
 
         # Execute cargo
         try:
+            # If quiet, capture all output and only show it in the exception
+            # If not quiet, forward all cargo output to stderr
             stderr = subprocess.PIPE if quiet else None
-            output = subprocess.check_output(command, env=env, stderr=stderr, text=True)
+            cargo_messages = subprocess.check_output(
+                command,
+                env=env,
+                stderr=stderr,
+                text=True,
+            )
         except subprocess.CalledProcessError as e:
-            raise CompileError(format_called_process_error(e))
+            # Don't include stdout in the formatted error as it is a huge dump
+            # of cargo json lines which aren't helpful for the end user.
+            raise CompileError(format_called_process_error(e, include_stdout=False))
 
         except OSError:
-            raise DistutilsExecError(
+            raise ExecError(
                 "Unable to execute 'cargo' - this package "
                 "requires Rust to be installed and cargo to be on the PATH"
             )
 
-        if not quiet:
-            if output:
-                print(output, file=sys.stderr)
-
         # Find the shared library that cargo hopefully produced and copy
         # it into the build directory as if it were produced by build_ext.
 
-        profile = ext.get_cargo_profile()
-        if profile:
-            # https://doc.rust-lang.org/cargo/reference/profiles.html
-            if profile in {"dev", "test"}:
-                profile_dir = "debug"
-            elif profile == "bench":
-                profile_dir = "release"
-            else:
-                profile_dir = profile
-        else:
-            profile_dir = "debug" if debug else "release"
-        artifacts_dir = os.path.join(target_dir, profile_dir)
         dylib_paths = []
 
         if ext._uses_exec_binding():
+            # Find artifact from cargo messages
+            artifacts = _find_cargo_artifacts(
+                cargo_messages.splitlines(),
+                package_id=package_id,
+                kinds={"bin"},
+            )
             for name, dest in ext.target.items():
                 if not name:
                     name = dest.split(".")[-1]
-                exe = sysconfig.get_config_var("EXE")
-                if exe is not None:
-                    name += exe
 
-                path = os.path.join(artifacts_dir, name)
-                if os.access(path, os.X_OK):
-                    dylib_paths.append(_BuiltModule(dest, path))
-                else:
-                    raise DistutilsExecError(
-                        "Rust build failed; "
-                        f"unable to find executable '{name}' in '{artifacts_dir}'"
+                try:
+                    artifact_path = next(
+                        artifact
+                        for artifact in artifacts
+                        if Path(artifact).with_suffix("").name == name
                     )
+                except StopIteration:
+                    raise ExecError(
+                        f"Rust build failed; unable to locate executable '{name}'"
+                    )
+
+                if os.environ.get("CARGO") == "cross":
+                    artifact_path = _replace_cross_target_dir(
+                        artifact_path, ext, quiet=quiet
+                    )
+
+                dylib_paths.append(_BuiltModule(dest, artifact_path))
         else:
-            platform = sysconfig.get_platform()
-            if "win" in platform:
-                dylib_ext = "dll"
-            elif platform.startswith("macosx"):
-                dylib_ext = "dylib"
-            elif "wasm32" in platform:
-                dylib_ext = "wasm"
-            else:
-                dylib_ext = "so"
-
-            wildcard_so = "*{}.{}".format(ext.get_lib_name(quiet=quiet), dylib_ext)
-
-            try:
-                dylib_paths.append(
-                    _BuiltModule(
-                        ext.name,
-                        next(glob.iglob(os.path.join(artifacts_dir, wildcard_so))),
-                    )
+            # Find artifact from cargo messages
+            artifacts = _find_cargo_artifacts(
+                cargo_messages.splitlines(),
+                package_id=package_id,
+                kinds={"cdylib", "dylib"},
+            )
+            if len(artifacts) == 0:
+                raise ExecError(
+                    "Rust build failed; unable to find any cdylib or dylib build artifacts"
                 )
-            except StopIteration:
-                raise DistutilsExecError(
-                    f"Rust build failed; unable to find any {wildcard_so} in {artifacts_dir}"
+            elif len(artifacts) > 1:
+                raise ExecError(
+                    f"Rust build failed; expected only one cdylib or dylib build artifact but found {artifacts}"
                 )
+
+            artifact_path = artifacts[0]
+
+            if os.environ.get("CARGO") == "cross":
+                artifact_path = _replace_cross_target_dir(
+                    artifact_path, ext, quiet=quiet
+                )
+
+            # guaranteed to be just one element after checks above
+            dylib_paths.append(_BuiltModule(ext.name, artifact_path))
         return dylib_paths
 
     def install_extension(
@@ -301,34 +344,64 @@ class build_rust(RustCommand):
                 )
 
             if ext._uses_exec_binding():
-                ext_path = build_ext.get_ext_fullpath(module_name)
-                # remove extensions
-                ext_path, _, _ = _split_platform_and_extension(ext_path)
-
-                # Add expected extension
                 exe = sysconfig.get_config_var("EXE")
-                if exe is not None:
-                    ext_path += exe
 
-                os.makedirs(os.path.dirname(ext_path), exist_ok=True)
                 if isinstance(ext, RustBin):
-                    executable_name = module_name
+                    # will install the rust binary into the scripts directory
+                    bin_name = module_name
                     if exe is not None:
-                        executable_name += exe
-                    wheel = self.get_finalized_command("bdist_wheel")
-                    scripts_dir = os.path.join(
-                        build_ext.build_lib, wheel.data_dir, "scripts"  # type: ignore[attr-defined]
+                        bin_name += exe
+
+                    install_scripts = cast(
+                        CommandInstallScripts,
+                        self.get_finalized_command("install_scripts"),
                     )
-                    os.makedirs(scripts_dir, exist_ok=True)
-                    ext_path = os.path.join(scripts_dir, executable_name)
+                    ext_path = os.path.join(install_scripts.build_dir, bin_name)
                 else:
+                    # will install the rust binary into the module directory
+                    ext_path = build_ext.get_ext_fullpath(module_name)
+
+                    # add expected extension
+                    ext_path, _, _ = _split_platform_and_extension(ext_path)
+                    if exe is not None:
+                        ext_path += exe
+
+                    # if required, also generate a console script entry point
                     ext.install_script(module_name.split(".")[-1], ext_path)
             else:
+                # will install the rust library into the module directory
                 ext_path = self.get_dylib_ext_path(ext, module_name)
-                os.makedirs(os.path.dirname(ext_path), exist_ok=True)
 
-            log.info("Copying rust artifact from %s to %s", dylib_path, ext_path)
-            shutil.copyfile(dylib_path, ext_path)
+            # Make filenames relative to cwd where possible, to make logs and
+            # errors below a little neater
+
+            cwd = os.getcwd()
+            if dylib_path.startswith(cwd):
+                dylib_path = os.path.relpath(dylib_path, cwd)
+            if ext_path.startswith(cwd):
+                ext_path = os.path.relpath(ext_path, cwd)
+
+            os.makedirs(os.path.dirname(ext_path), exist_ok=True)
+            logger.info("Copying rust artifact from %s to %s", dylib_path, ext_path)
+
+            # We want to atomically replace any existing library file. We can't
+            # just copy the new library directly on top of the old one as that
+            # causes the existing library to be modified (rather the replaced).
+            # This means that any process that currently uses the shared library
+            # will see it modified and likely segfault.
+            #
+            # We first copy the file to the same directory, as `os.replace`
+            # doesn't work across file system boundaries.
+            temp_ext_path = ext_path + "~"
+            shutil.copyfile(dylib_path, temp_ext_path)
+            try:
+                os.replace(temp_ext_path, ext_path)
+            except PermissionError as e:
+                msg = f"{e}\n  hint: check permissions for {ext_path!r}"
+                if sys.platform == "win32":
+                    # On Windows, dll files are locked by the system when in use.
+                    msg += "\n  hint: the file may be in use by another Python process"
+                raise CompileError(msg)
 
             if sys.platform != "win32" and not debug_build:
                 args = []
@@ -341,7 +414,7 @@ class build_rust(RustCommand):
                     args.insert(0, "strip")
                     args.append(ext_path)
                     try:
-                        output = subprocess.check_output(args)
+                        subprocess.check_output(args)
                     except subprocess.CalledProcessError:
                         pass
 
@@ -392,59 +465,22 @@ class build_rust(RustCommand):
         return ext_path
 
     def _py_limited_api(self) -> _PyLimitedApi:
-        bdist_wheel = self.distribution.get_command_obj("bdist_wheel", create=False)
+        bdist_wheel = _get_bdist_wheel_cmd(self.distribution, create=False)
 
         if bdist_wheel is None:
             # wheel package is not installed, not building a limited-api wheel
             return False
         else:
-            from wheel.bdist_wheel import bdist_wheel as CommandBdistWheel
-
-            bdist_wheel_command = cast(CommandBdistWheel, bdist_wheel)  # type: ignore[no-any-unimported]
-            bdist_wheel_command.ensure_finalized()
-            return cast(_PyLimitedApi, bdist_wheel_command.py_limited_api)
+            return cast(_PyLimitedApi, bdist_wheel.py_limited_api)
 
     def _detect_rust_target(
         self, forced_target_triple: Optional[str] = None
-    ) -> Optional["_TargetInfo"]:
+    ) -> Optional[str]:
         assert self.plat_name is not None
-        cross_compile_info = _detect_unix_cross_compile_info()
-        if cross_compile_info is not None:
-            cross_target_info = cross_compile_info.to_target_info()
-            if forced_target_triple is not None:
-                if (
-                    cross_target_info is not None
-                    and not cross_target_info.is_compatible_with(forced_target_triple)
-                ):
-                    self.warn(
-                        f"Forced Rust target `{forced_target_triple}` is not "
-                        f"compatible with deduced Rust target "
-                        f"`{cross_target_info.triple}` - the built package "
-                        f" may not import successfully once installed."
-                    )
-
-                # Forcing the target in a cross-compile environment; use
-                # the cross-compile information in combination with the
-                # forced target
-                return _TargetInfo(
-                    forced_target_triple,
-                    cross_compile_info.cross_lib,
-                    cross_compile_info.linker,
-                    cross_compile_info.linker_args,
-                )
-            elif cross_target_info is not None:
-                return cross_target_info
-            else:
-                raise DistutilsPlatformError(
-                    "Don't know the correct rust target for system type "
-                    f"{cross_compile_info.host_type}. Please set the "
-                    "CARGO_BUILD_TARGET environment variable."
-                )
-
-        elif forced_target_triple is not None:
+        if forced_target_triple is not None:
             # Automatic target detection can be overridden via the CARGO_BUILD_TARGET
             # environment variable or --target command line option
-            return _TargetInfo.for_triple(forced_target_triple)
+            return forced_target_triple
 
         # Determine local rust target which needs to be "forced" if necessary
         local_rust_target = _adjusted_local_rust_target(self.plat_name)
@@ -456,7 +492,7 @@ class build_rust(RustCommand):
             # check for None first to avoid calling to rustc if not needed
             and local_rust_target != get_rust_host()
         ):
-            return _TargetInfo.for_triple(local_rust_target)
+            return local_rust_target
 
         return None
 
@@ -481,10 +517,10 @@ class build_rust(RustCommand):
         if target_triple is not None:
             args.extend(["--target", target_triple])
 
-        if release:
-            profile = ext.get_cargo_profile()
-            if not profile:
-                args.append("--release")
+        ext_profile = ext.get_cargo_profile()
+        env_profile = os.getenv("SETUPTOOLS_RUST_CARGO_PROFILE")
+        if release and not ext_profile and not env_profile:
+            args.append("--release")
 
         if quiet:
             args.append("-q")
@@ -505,6 +541,18 @@ class build_rust(RustCommand):
         if ext.args is not None:
             args.extend(ext.args)
 
+        if env_profile:
+            if ext_profile:
+                args = [p for p in args if not p.startswith("--profile=")]
+                while True:
+                    try:
+                        index = args.index("--profile")
+                        del args[index : index + 2]
+                    except ValueError:
+                        break
+
+            args.extend(["--profile", env_profile])
+
         if ext.cargo_manifest_args is not None:
             args.extend(ext.cargo_manifest_args)
 
@@ -524,7 +572,7 @@ def create_universal2_binary(output_path: str, input_paths: List[str]) -> None:
         try:
             from fat_macho import FatWriter
         except ImportError:
-            raise DistutilsExecError(
+            raise ExecError(
                 "failed to locate `lipo` or import `fat_macho.FatWriter`. "
                 "Try installing with `pip install fat-macho` "
             )
@@ -546,91 +594,6 @@ class _BuiltModule(NamedTuple):
     path: str
 
 
-class _TargetInfo(NamedTuple):
-    triple: str
-    cross_lib: Optional[str]
-    linker: Optional[str]
-    linker_args: Optional[str]
-
-    @staticmethod
-    def for_triple(triple: str) -> "_TargetInfo":
-        return _TargetInfo(triple, None, None, None)
-
-    def is_compatible_with(self, target: str) -> bool:
-        if self.triple == target:
-            return True
-
-        # the vendor field can be ignored, so x86_64-pc-linux-gnu is compatible
-        # with x86_64-unknown-linux-gnu
-        if _replace_vendor_with_unknown(self.triple) == target:
-            return True
-
-        return False
-
-
-class _CrossCompileInfo(NamedTuple):
-    host_type: str
-    cross_lib: Optional[str]
-    linker: Optional[str]
-    linker_args: Optional[str]
-
-    def to_target_info(self) -> Optional[_TargetInfo]:
-        """Maps this cross compile info to target info.
-
-        Returns None if the corresponding target information could not be
-        deduced.
-        """
-        # hopefully an exact match
-        targets = get_rust_target_list()
-        if self.host_type in targets:
-            return _TargetInfo(
-                self.host_type, self.cross_lib, self.linker, self.linker_args
-            )
-
-        # the vendor field can be ignored, so x86_64-pc-linux-gnu is compatible
-        # with x86_64-unknown-linux-gnu
-        without_vendor = _replace_vendor_with_unknown(self.host_type)
-        if without_vendor is not None and without_vendor in targets:
-            return _TargetInfo(
-                without_vendor, self.cross_lib, self.linker, self.linker_args
-            )
-
-        return None
-
-
-def _detect_unix_cross_compile_info() -> Optional["_CrossCompileInfo"]:
-    # See https://github.com/PyO3/setuptools-rust/issues/138
-    # This is to support cross compiling on *NIX, where plat_name isn't
-    # necessarily the same as the system we are running on.  *NIX systems
-    # have more detailed information available in sysconfig. We need that
-    # because plat_name doesn't give us information on e.g., glibc vs musl.
-    host_type = sysconfig.get_config_var("HOST_GNU_TYPE")
-    build_type = sysconfig.get_config_var("BUILD_GNU_TYPE")
-
-    if not host_type or host_type == build_type:
-        # not *NIX, or not cross compiling
-        return None
-
-    if "apple-darwin" in host_type and (build_type and "apple-darwin" in build_type):
-        # On macos and the build and host differ. This is probably an arm
-        # Python which was built on x86_64. Don't try to handle this for now.
-        # (See https://github.com/PyO3/setuptools-rust/issues/192)
-        return None
-
-    stdlib = sysconfig.get_path("stdlib")
-    assert stdlib is not None
-    cross_lib = os.path.dirname(stdlib)
-
-    bldshared = sysconfig.get_config_var("BLDSHARED")
-    if not bldshared:
-        linker = None
-        linker_args = None
-    else:
-        [linker, _, linker_args] = bldshared.partition(" ")
-
-    return _CrossCompileInfo(host_type, cross_lib, linker, linker_args)
-
-
 def _replace_vendor_with_unknown(target: str) -> Optional[str]:
     """Replaces vendor in the target triple with unknown.
 
@@ -643,7 +606,7 @@ def _replace_vendor_with_unknown(target: str) -> Optional[str]:
     return "-".join(components)
 
 
-def _prepare_build_environment(cross_lib: Optional[str]) -> Dict[str, str]:
+def _prepare_build_environment() -> Dict[str, str]:
     """Prepares environment variables to use when executing cargo build."""
 
     # Make sure that if pythonXX-sys is used, it builds against the current
@@ -663,27 +626,9 @@ def _prepare_build_environment(cross_lib: Optional[str]) -> Dict[str, str]:
             "PYO3_PYTHON": os.environ.get("PYO3_PYTHON", sys.executable),
         }
     )
-
-    if cross_lib:
-        env.setdefault("PYO3_CROSS_LIB_DIR", cross_lib)
-
-    env.pop("CARGO", None)
     return env
 
 
-def _base_cargo_target_dir(ext: RustExtension, *, quiet: bool) -> str:
-    """Returns the root target directory cargo will use.
-
-    If --target is passed to cargo in the command line, the target directory
-    will have the target appended as a child.
-    """
-    target_directory = ext._metadata(quiet=quiet)["target_directory"]
-    assert isinstance(
-        target_directory, str
-    ), "expected cargo metadata to contain a string target directory"
-    return target_directory
-
-
 def _is_py_limited_api(
     ext_setting: Literal["auto", True, False],
     wheel_setting: Optional[_PyLimitedApi],
@@ -724,12 +669,12 @@ def _binding_features(
                 python_version = py_limited_api[2:]
                 features.add(f"pyo3/abi3-py{python_version}")
             elif py_limited_api:
-                features.add(f"pyo3/abi3")
+                features.add("pyo3/abi3")
         return features
     elif ext.binding is Binding.RustCPython:
         return {"cpython/python3-sys", "cpython/extension-module"}
     else:
-        raise DistutilsPlatformError(f"unknown Rust binding: '{ext.binding}'")
+        raise PlatformError(f"unknown Rust binding: '{ext.binding}'")
 
 
 _PyLimitedApi = Literal["cp37", "cp38", "cp39", "cp310", "cp311", "cp312", True, False]
@@ -770,3 +715,77 @@ def _split_platform_and_extension(ext_path: str) -> Tuple[str, str, str]:
     # rust.cpython-38-x86_64-linux-gnu to (rust, .cpython-38-x86_64-linux-gnu)
     ext_path, platform_tag = os.path.splitext(ext_path)
     return (ext_path, platform_tag, extension)
+
+
+def _find_cargo_artifacts(
+    cargo_messages: List[str],
+    *,
+    package_id: str,
+    kinds: Set[str],
+) -> List[str]:
+    """Identifies cargo artifacts built for the given `package_id` from the
+    provided cargo_messages.
+
+    >>> _find_cargo_artifacts(
+    ...    [
+    ...        '{"some_irrelevant_message": []}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib"]},"filenames":["/some/path/baz.so"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["dylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
+    ...    ],
+    ...    package_id="some_id",
+    ...    kinds={"cdylib", "dylib"},
+    ... )
+    ['/some/path/baz.so', '/file/two/baz.dylib']
+    >>> _find_cargo_artifacts(
+    ...    [
+    ...        '{"some_irrelevant_message": []}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib"]},"filenames":["/some/path/baz.so"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
+    ...    ],
+    ...    package_id="some_id",
+    ...    kinds={"rlib"},
+    ... )
+    ['/file/two/baz.rlib']
+    """
+    artifacts = []
+    for message in cargo_messages:
+        # only bother parsing messages that look like a match
+        if "compiler-artifact" in message and package_id in message:
+            parsed = json.loads(message)
+            # verify the message is correct
+            if (
+                parsed.get("reason") == "compiler-artifact"
+                and parsed.get("package_id") == package_id
+            ):
+                for artifact_kind, filename in zip(
+                    parsed["target"]["kind"], parsed["filenames"]
+                ):
+                    if artifact_kind in kinds:
+                        artifacts.append(filename)
+    return artifacts
+
+
+def _replace_cross_target_dir(path: str, ext: RustExtension, *, quiet: bool) -> str:
+    """Replaces target director from `cross` docker build with the correct
+    local path.
+
+    Cross artifact messages and metadata contain paths from inside the
+    dockerfile; invoking `cargo metadata` we can work out the correct local
+    target directory.
+    """
+    cross_target_dir = ext._metadata(cargo="cross", quiet=quiet)["target_directory"]
+    local_target_dir = ext._metadata(cargo="cargo", quiet=quiet)["target_directory"]
+    return path.replace(cross_target_dir, local_target_dir)
+
+
+def _get_bdist_wheel_cmd(  # type: ignore[no-any-unimported]
+    dist: Distribution, create: bool = True
+) -> Optional[CommandBdistWheel]:
+    try:
+        cmd_obj = dist.get_command_obj("bdist_wheel", create=create)
+        cmd_obj.ensure_finalized()  # type: ignore[union-attr]
+        return cast(CommandBdistWheel, cmd_obj)  # type: ignore[no-any-unimported]
+    except Exception:
+        return None
